// Subdivider.cpp: Subdivider NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Subdivider.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

Subdivider::Subdivider()
{
	vertex_index = NULL;
}

Subdivider::~Subdivider()
{
	if(vertex_index != NULL)
		delete[] vertex_index;
}

void Subdivider::setMeshData(MeshData *mesh)
{
	this->mesh = mesh;
}

void Subdivider::linear()
{
	int (*face)[3] = mesh->face;
	float (*vertex)[3] = mesh->vertex;
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*link)[3] = mesh->face_link_E;

	BOOL* flag = new BOOL[face_N];
	for(int i=0; i<face_N; i++){
		//if(vertex[face[i][0]][2] > 0)
			flag[i] = true;
		//else
			//flag[i] = false;
	}

	this->subdivideTopology(flag);

	for(i=0; i<vertex_N; i++){
		new_vertex[i][0] = vertex[i][0];
		new_vertex[i][1] = vertex[i][1];
		new_vertex[i][2] = vertex[i][2];
	}

	int counter= 0;
	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link[i][j];
			if(pair < i || !flag[pair]){
				int index = vertex_index[i][j];
				int i1 = face[i][j];
				int i2 = face[i][(j+1)%3];
				new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
				new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
				new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);		
			}
		}
	}

	delete[] flag;
	delete[] vertex_index;
	vertex_index = NULL;

	this->updateMesh();	
}

void Subdivider::LOOP()
{
	int (*face)[3] = mesh->face;
	float (*vertex)[3] = mesh->vertex;
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*link_f)[3] = mesh->face_link_E;
	BOOL* isBound = mesh->isBound;

	BOOL* flag = new BOOL[face_N];
	for(int i=0; i<face_N; i++)
		flag[i] = true;

	this->subdivideTopology(flag);

	for(i=0; i<vertex_N; i++){
		if(isBound[i] == INNER){
			int *link, link_N;
			mesh->getConsistent1Ring(i, link, link_N);
			if(link_N == 0){
				new_vertex[i][0] = vertex[i][0];
				new_vertex[i][1] = vertex[i][1];
				new_vertex[i][2] = vertex[i][2];
				continue;
			}
			double b = (0.625 - (0.375 + 0.25*cos(2.0*PI/(double)link_N))
				*(0.375 + 0.25*cos(2.0*PI/(double)link_N)))/(double)link_N;
			double sum[3];
			sum[0] = sum[1] = sum[2] = 0;
			for(int j=0; j<link_N; j++){
				sum[0] += vertex[link[j]][0];
				sum[1] += vertex[link[j]][1];
				sum[2] += vertex[link[j]][2];
			}
			new_vertex[i][0] = (float)((1.0-(double)link_N*b)*vertex[i][0] + b*sum[0]);
			new_vertex[i][1] = (float)((1.0-(double)link_N*b)*vertex[i][1] + b*sum[1]);
			new_vertex[i][2] = (float)((1.0-(double)link_N*b)*vertex[i][2] + b*sum[2]);
		}
		else if(isBound[i] == BOUNDARY){
			int *link, link_N;
			mesh->getConsistent1Ring(i, link, link_N);
			int i1 = link[0];
			int i2 = link[link_N-1];
			new_vertex[i][0] = (float)(0.125*(vertex[i1][0] + vertex[i2][0] + 6.0*vertex[i][0]));
			new_vertex[i][1] = (float)(0.125*(vertex[i1][1] + vertex[i2][1] + 6.0*vertex[i][1]));
			new_vertex[i][2] = (float)(0.125*(vertex[i1][2] + vertex[i2][2] + 6.0*vertex[i][2]));
		}
		else{
			new_vertex[i][0] = vertex[i][0];
			new_vertex[i][1] = vertex[i][1];
			new_vertex[i][2] = vertex[i][2];
		}
	}

	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link_f[i][j];
			if(pair < i || !flag[pair]){
				int index = vertex_index[i][j];
				int i1 = face[i][j];
				int i2 = face[i][(j+1)%3];

				if(pair < 0){
					new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
					new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
					new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);
					continue;
				}
			
				int i3 = face[i][(j+2)%3];
				int i4;
				if(link_f[pair][0] == i)
					i4 = face[pair][2];
				else if(link_f[pair][1] == i)
					i4 = face[pair][0];
				else
					i4 = face[pair][1];
				if(isBound[i1] == INNER || isBound[i2] == INNER ||
					(i3 >= 0 && i4 >= 0)){
					new_vertex[index][0] = (float)(0.125*(3.0*(vertex[i1][0] + vertex[i2][0]) 
														+ vertex[i3][0] + vertex[i4][0]));
					new_vertex[index][1] = (float)(0.125*(3.0*(vertex[i1][1] + vertex[i2][1]) 
														+ vertex[i3][1] + vertex[i4][1]));
					new_vertex[index][2] = (float)(0.125*(3.0*(vertex[i1][2] + vertex[i2][2]) 
														+ vertex[i3][2] + vertex[i4][2]));
				}
				else{
					new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
					new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
					new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);
				}
			}
		}
	}

	delete[] flag;
	delete[] vertex_index;
	vertex_index = NULL;
	
	this->updateMesh();

	
}

void Subdivider::subdivideTopology(BOOL *flag)
{
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*link)[3] = mesh->face_link_E;
	float (*vertex)[3] = mesh->vertex;
	int (*face)[3] = mesh->face;
	BOOL grow = true;
	while(grow){
		grow = false;
		for(int i=0; i<face_N; i++){
			if(flag[i])
				continue;
			int count = 0;
			for(int j=0; j<3; j++){
				if(link[i][j] >= 0 && flag[link[i][j]])
					count++;
			}
			if(count > 1){
				flag[i] = true;
				grow = true;
			}
		}
	}

	if(vertex_index != NULL)
		delete[] vertex_index;
	vertex_index = new int[face_N][3];
	for(int i=0; i<face_N; i++)
		vertex_index[i][0] = vertex_index[i][1] = vertex_index[i][2] = -1;

	new_vertex_N = vertex_N;
	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link[i][j];
			if(pair < i || !flag[pair]){
				vertex_index[i][j] = new_vertex_N;
				if(pair >= 0){
					if(link[pair][0] == i)
						vertex_index[pair][0] = new_vertex_N;
					else if(link[pair][1] == i)
						vertex_index[pair][1] = new_vertex_N;
					else
						vertex_index[pair][2] = new_vertex_N;
				}
				new_vertex_N++;
			}
		}
	}

	new_vertex = new float[new_vertex_N][3];

	new_face_N = face_N;
	for(i=0; i<face_N; i++){
		if(flag[i])
			new_face_N += 3;
		else{
			int count = 0;
			for(int j=0; j<3; j++){
				if(flag[link[i][j]])
					count++;
			}
			if(count != 0)
				new_face_N++;
		}
	}
	new_face = new int[new_face_N][3];
	int face_id = 0;
	for(i=0; i<face_N; i++){
		if(flag[i]){
			new_face[face_id][0] = face[i][0];
			new_face[face_id][1] = vertex_index[i][0];
			new_face[face_id][2] = vertex_index[i][2];
			face_id++;

			new_face[face_id][0] = face[i][1];
			new_face[face_id][1] = vertex_index[i][1];
			new_face[face_id][2] = vertex_index[i][0];
			face_id++;

			new_face[face_id][0] = face[i][2];
			new_face[face_id][1] = vertex_index[i][2];
			new_face[face_id][2] = vertex_index[i][1];
			face_id++;

			new_face[face_id][0] = vertex_index[i][0];
			new_face[face_id][1] = vertex_index[i][1];
			new_face[face_id][2] = vertex_index[i][2];
			face_id++;
		}
		else{
			int id = -1;
			for(int j=0; j<3; j++){
				if(link[i][j] >= 0 && flag[link[i][j]])
					id = j;
			}
			switch(id){
			case -1:
				new_face[face_id][0] = face[i][0];
				new_face[face_id][1] = face[i][1];
				new_face[face_id][2] = face[i][2];
				face_id++;
				break;

			case 0:
				new_face[face_id][0] = face[i][0];
				new_face[face_id][1] = vertex_index[i][0];
				new_face[face_id][2] = face[i][2];
				face_id++;

				new_face[face_id][0] = face[i][1];
				new_face[face_id][1] = face[i][2];
				new_face[face_id][2] = vertex_index[i][0];
				face_id++;
				break;

			case 1:
				new_face[face_id][0] = face[i][0];
				new_face[face_id][1] = face[i][1];
				new_face[face_id][2] = vertex_index[i][1];
				face_id++;

				new_face[face_id][0] = face[i][2];
				new_face[face_id][1] = face[i][0];
				new_face[face_id][2] = vertex_index[i][1];
				face_id++;
				break;

			case 2:
				new_face[face_id][0] = face[i][0];
				new_face[face_id][1] = face[i][1];
				new_face[face_id][2] = vertex_index[i][2];
				face_id++;

				new_face[face_id][0] = face[i][1];
				new_face[face_id][1] = face[i][2];
				new_face[face_id][2] = vertex_index[i][2];
				face_id++;
				break;
			}	
		}
	}
}

void Subdivider::updateMesh()
{
	bool* new_fix = new bool[new_vertex_N];
	for(int i=0; i<mesh->vertex_N; i++)
		new_fix[i] = mesh->isFixed[i];
	for(i=mesh->vertex_N; i<new_vertex_N; i++)
		new_fix[i] = false;

	mesh->deleteLinkData();
	mesh->vertex_N = new_vertex_N;
	mesh->vertex = new_vertex;
	mesh->face_N = new_face_N;
	mesh->face = new_face;
	
	

	mesh->isFixed = new_fix;
	mesh->generateVertexLink();
	mesh->generateFaceLink(); 
	mesh->computeFaceNormal();
	mesh->computeNormal();
}

void Subdivider::butterfly()
{
	int (*face)[3] = mesh->face;
	float (*vertex)[3] = mesh->vertex;
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*link_f)[3] = mesh->face_link_E;
	BOOL* isBound = mesh->isBound;
	int* degree = mesh->degree_v;

	BOOL* flag = new BOOL[face_N];
	for(int i=0; i<face_N; i++){
		//if(vertex[face[i][0]][2] > 0)
			flag[i] = true;
		//else
			//flag[i] = false;
	}

	this->subdivideTopology(flag);

	for(i=0; i<vertex_N; i++){
		new_vertex[i][0] = vertex[i][0];
		new_vertex[i][1] = vertex[i][1];
		new_vertex[i][2] = vertex[i][2];
	}

	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link_f[i][j];
			if(pair < i || !flag[pair]){
				int index = vertex_index[i][j];
				int i1 = face[i][j];
				int i2 = face[i][(j+1)%3];

				int *link1, *link2, link1_N, link2_N;
				mesh->getConsistent1Ring(i1, link1, link1_N);
				mesh->getConsistent1Ring(i2, link2, link2_N);

				if(pair < 0 && isBound[i1] == BOUNDARY && isBound[i2] == BOUNDARY){
					int i3 = link1[0];
					if(i3 == i2)
						i3 = link1[link1_N-1];
					int i4 = link2[0];
					if(i4 == i1)
						i4 = link2[link2_N-1];
					new_vertex[index][0] = (float)(0.0625*(9.0*(vertex[i1][0] + vertex[i2][0]) 
						- (vertex[i3][0] + vertex[i4][0])));
					new_vertex[index][1] = (float)(0.0625*(9.0*(vertex[i1][1] + vertex[i2][1]) 
						- (vertex[i3][1] + vertex[i4][1])));
					new_vertex[index][2] = (float)(0.0625*(9.0*(vertex[i1][2] + vertex[i2][2]) 
						- (vertex[i3][2] + vertex[i4][2])));
					continue;
				}

				int i3, i4;
				i3 = face[i][(j+2)%3];
				if(link_f[pair][0] == i)
					i4 = face[pair][2];
				else if(link_f[pair][1] == i)
					i4 = face[pair][0];
				else
					i4 = face[pair][1];

				if(isBound[i1] == INNER && isBound[i2] == INNER){
					if(degree[i1] == 6 && degree[i2] == 6){
						int index1, index2;
						for(int k=0; k<link1_N; k++)
							if(link1[k] == i2)
								index1 = k;
						for(k=0; k<link2_N; k++)
							if(link2[k] == i1)
								index2 = k;
						int i5, i6, i7, i8;
						i5 = link2[(index2+link1_N-2)%link2_N];
						i6 = link1[(index1+2)%link1_N];
						i7 = link1[(index1+link1_N-2)%link1_N];
						i8 = link2[(index2+2)%link2_N];

						new_vertex[index][0] = (float)(0.0625*(8.0*(vertex[i1][0] + vertex[i2][0]) 
							+ 2.0*(vertex[i3][0] + vertex[i4][0])
							- (vertex[i5][0] + vertex[i6][0] + vertex[i7][0] + vertex[i8][0])));
						new_vertex[index][1] = (float)(0.0625*(8.0*(vertex[i1][1] + vertex[i2][1]) 
							+ 2.0*(vertex[i3][1] + vertex[i4][1])
							- (vertex[i5][1] + vertex[i6][1] + vertex[i7][1] + vertex[i8][1])));
						new_vertex[index][2] = (float)(0.0625*(8.0*(vertex[i1][2] + vertex[i2][2]) 
							+ 2.0*(vertex[i3][2] + vertex[i4][2])
							- (vertex[i5][2] + vertex[i6][2] + vertex[i7][2] + vertex[i8][2])));
					}
					else{
						if(degree[i1] == 6 && degree[i2] != 6){
							int tmp = i1;
							i1 = i2;
							i2 = tmp;
							link1 = link2;
							link1_N = link2_N;
						}
						if(degree[i1] != 6 && degree[i2] == 6){
							int index1;
							for(int k=0; k<link1_N; k++){
								if(link1[k] == i2)
									index1 = k;
							}
							double sum[3];
							
							if(degree[i1] == 3){
								sum[0] = 0.25*(5.0*vertex[i2][0] 
									- (vertex[link1[(index1+1)%3]][0] + vertex[link1[(index1+2)%3]][0]));
								sum[1] = 0.25*(5.0*vertex[i2][1] 
									- (vertex[link1[(index1+1)%3]][1] + vertex[link1[(index1+2)%3]][1]));
								sum[2] = 0.25*(5.0*vertex[i2][2] 
									- (vertex[link1[(index1+1)%3]][2] + vertex[link1[(index1+2)%3]][2]));
							}
							else if(degree[i1] == 4){
								sum[0] = 0.5*(3.0*vertex[i2][0] - vertex[link1[(index1+2)%4]][0]);
								sum[1] = 0.5*(3.0*vertex[i2][1] - vertex[link1[(index1+2)%4]][1]);
								sum[2] = 0.5*(3.0*vertex[i2][2] - vertex[link1[(index1+2)%4]][2]);
							}
							else{
								sum[0] = sum[1] = sum[2] = 0;
								for(int l=0; l<link1_N; l++){
									int m = link1[(index1+l)%link1_N];
									double s = (0.25 + cos(2.0*PI*(double)l/(double)link1_N) 
										+ 0.5*cos(4.0*PI*(double)l/(double)link1_N));
									sum[0] += s*vertex[m][0];
									sum[1] += s*vertex[m][1];
									sum[2] += s*vertex[m][2];
								}
							}
							new_vertex[index][0] = (float)(0.75*vertex[i1][0] + sum[0]/(double)link1_N);
							new_vertex[index][1] = (float)(0.75*vertex[i1][1] + sum[1]/(double)link1_N);
							new_vertex[index][2] = (float)(0.75*vertex[i1][2] + sum[2]/(double)link1_N);
						}
						else{
							int index1;
							for(int k=0; k<link1_N; k++){
								if(link1[k] == i2)
									index1 = k;
							}
							double sum[3];
							
							if(degree[i1] == 3){
								sum[0] = 0.25*(5.0*vertex[i2][0] 
									- (vertex[link1[(index1+1)%3]][0] + vertex[link1[(index1+2)%3]][0]));
								sum[1] = 0.25*(5.0*vertex[i2][1] 
									- (vertex[link1[(index1+1)%3]][1] + vertex[link1[(index1+2)%3]][1]));
								sum[2] = 0.25*(5.0*vertex[i2][2] 
									- (vertex[link1[(index1+1)%3]][2] + vertex[link1[(index1+2)%3]][2]));
							}
							else if(degree[i1] == 4){
								sum[0] = 0.5*(3.0*vertex[i2][0] - vertex[link1[(index1+2)%4]][0]);
								sum[1] = 0.5*(3.0*vertex[i2][1] - vertex[link1[(index1+2)%4]][1]);
								sum[2] = 0.5*(3.0*vertex[i2][2] - vertex[link1[(index1+2)%4]][2]);
							}
							else{
								sum[0] = sum[1] = sum[2] = 0;
								for(int l=0; l<link1_N; l++){
									int m = link1[(index1+l)%link1_N];
									double s = (0.25 + cos(2.0*PI*(double)l/(double)link1_N) 
										+ 0.5*cos(4.0*PI*(double)l/(double)link1_N));
									sum[0] += s*vertex[m][0];
									sum[1] += s*vertex[m][1];
									sum[2] += s*vertex[m][2];
								}
							}
							double tmp1[3];
							tmp1[0] = 0.75*vertex[i1][0] + sum[0]/(double)link1_N;
							tmp1[1] = 0.75*vertex[i1][1] + sum[1]/(double)link1_N;
							tmp1[2] = 0.75*vertex[i1][2] + sum[2]/(double)link1_N;

							int tmp = i1;
							i1 = i2;
							i2 = tmp;
							link1 = link2;
							link1_N = link2_N;

							for(k=0; k<link1_N; k++){
								if(link1[k] == i2)
									index1 = k;
							}
							
							if(degree[i1] == 3){
								sum[0] = 0.25*(5.0*vertex[i2][0] 
									- (vertex[link1[(index1+1)%3]][0] + vertex[link1[(index1+2)%3]][0]));
								sum[1] = 0.25*(5.0*vertex[i2][1] 
									- (vertex[link1[(index1+1)%3]][1] + vertex[link1[(index1+2)%3]][1]));
								sum[2] = 0.25*(5.0*vertex[i2][2] 
									- (vertex[link1[(index1+1)%3]][2] + vertex[link1[(index1+2)%3]][2]));
							}
							else if(degree[i1] == 4){
								sum[0] = 0.5*(3.0*vertex[i2][0] - vertex[link1[(index1+2)%4]][0]);
								sum[1] = 0.5*(3.0*vertex[i2][1] - vertex[link1[(index1+2)%4]][1]);
								sum[2] = 0.5*(3.0*vertex[i2][2] - vertex[link1[(index1+2)%4]][2]);
							}
							else{
								sum[0] = sum[1] = sum[2] = 0;
								for(int l=0; l<link1_N; l++){
									int m = link1[(index1+l)%link1_N];
									double s = (0.25 + cos(2.0*PI*(double)l/(double)link1_N) 
										+ 0.5*cos(4.0*PI*(double)l/(double)link1_N));
									sum[0] += s*vertex[m][0];
									sum[1] += s*vertex[m][1];
									sum[2] += s*vertex[m][2];
								}
							}
							double tmp2[3];
							tmp2[0] = 0.75*vertex[i1][0] + sum[0]/(double)link1_N;
							tmp2[1] = 0.75*vertex[i1][1] + sum[1]/(double)link1_N;
							tmp2[2] = 0.75*vertex[i1][2] + sum[2]/(double)link1_N;

							new_vertex[index][0] = (float)(0.5*(tmp1[0] + tmp2[0]));
							new_vertex[index][1] = (float)(0.5*(tmp1[1] + tmp2[1]));
							new_vertex[index][2] = (float)(0.5*(tmp1[2] + tmp2[2]));
						}
					}
				}
				else{
					new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
					new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
					new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);
				}
			}
		}
	}

	delete[] flag;
	delete[] vertex_index;
	vertex_index = NULL;
	
	this->updateMesh();

	
}

void Subdivider::subdivideCurvedFace(vtkImplicitFunction *func, float T)
{
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*face)[3] = mesh->face;
	BOOL* flag = new BOOL[face_N];
	float (*vertex)[3] = mesh->vertex;
	mesh->computeFaceNormal();
	float (*normal)[3] = mesh->normal_f;
	int (*link)[3] = mesh->face_link_E;

	for(int i=0; i<face_N; i++){
		if(face[i][0] < 0){
			flag[i] = false;
			continue;
		}
		float err = 0;
		double a = mesh->faceArea(i);
		double c[3];
		mesh->faceCenter(c, i);
		float v[3][3];
		for(int j=0; j<3; j++){
			v[j][0] = vertex[face[i][j]][0];
			v[j][1] = vertex[face[i][j]][1];
			v[j][2] = vertex[face[i][j]][2];
		}
		float c1[3];
		c1[0] = (float)c[0];
		c1[1] = (float)c[1];
		c1[2] = (float)c[2];
		float g[3];
		func->EvaluateGradient(c1, g);
		
		err += (float)(1.0-fabs(MeshData::DOT(normal[i], g))/MeshData::LENGTH(g));
		
		if(a*err > T){
			flag[i] = true;
			continue;
		}

		for(j=0; j<3; j++){
			c1[0] = (4.0f*v[(j+1)%3][0] + v[(j+2)%3][0] + v[j][0])/6.0f;
			c1[1] = (4.0f*v[(j+1)%3][1] + v[(j+2)%3][1] + v[j][1])/6.0f;
			c1[2] = (4.0f*v[(j+1)%3][2] + v[(j+2)%3][2] + v[j][2])/6.0f;

			func->EvaluateGradient(c1, g);
		
			err += (float)(1.0-fabs(MeshData::DOT(normal[i], g))/MeshData::LENGTH(g));
		}

		if(0.25*a*err > T)
			flag[i] = true;
		else
			flag[i] = false;
	}

	this->subdivideTopology(flag);

	for(i=0; i<vertex_N; i++){
		new_vertex[i][0] = vertex[i][0];
		new_vertex[i][1] = vertex[i][1];
		new_vertex[i][2] = vertex[i][2];
	}

	int counter= 0;
	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link[i][j];
			if(pair < i || !flag[pair]){
				int index = vertex_index[i][j];
				int i1 = face[i][j];
				int i2 = face[i][(j+1)%3];
				new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
				new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
				new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);		
			}
		}
	}

	delete[] flag;
	delete[] vertex_index;
	vertex_index = NULL;

	this->updateMesh();
}

void Subdivider::subdivideInside(vtkImplicitFunction *func)
{
	int face_N = mesh->face_N;
	int vertex_N = mesh->vertex_N;
	int (*face)[3] = mesh->face;
	BOOL* flag = new BOOL[face_N];
	float (*vertex)[3] = mesh->vertex;
	//mesh->computeFaceNormal();
	//float (*normal)[3] = mesh->normal_f;
	int (*link)[3] = mesh->face_link_E;

	for(int i=0; i<face_N; i++){
		flag[i] = false;
		for(int j=0; j<3; j++){
			if(func->EvaluateFunction(vertex[face[i][j]]) >= 0){
				flag[i] = true;
				break;
			}
		}
	}

	this->subdivideTopology(flag);

	for(i=0; i<vertex_N; i++){
		new_vertex[i][0] = vertex[i][0];
		new_vertex[i][1] = vertex[i][1];
		new_vertex[i][2] = vertex[i][2];
	}

	int counter= 0;
	for(i=0; i<face_N; i++){
		if(!flag[i])
			continue;
		for(int j=0; j<3; j++){
			int pair = link[i][j];
			if(pair < i || !flag[pair]){
				int index = vertex_index[i][j];
				int i1 = face[i][j];
				int i2 = face[i][(j+1)%3];
				new_vertex[index][0] = 0.5f*(vertex[i1][0] + vertex[i2][0]);
				new_vertex[index][1] = 0.5f*(vertex[i1][1] + vertex[i2][1]);
				new_vertex[index][2] = 0.5f*(vertex[i1][2] + vertex[i2][2]);		
			}
		}
	}

	delete[] flag;
	delete[] vertex_index;
	vertex_index = NULL;

	this->updateMesh();
}
